use bitboard::*;
use movegen::*;
use game::Game;
use evaluate::*;
use log::info;
use rand::prelude::*;

enum MoveUndo {
    Default {
        piece_type: PieceType,
        before: Bitboard,
        captured_type: PieceType,
        captured_before: Bitboard,
    },
    Castling { rooks_before: Bitboard, king_before: Bitboard },
    Promotion {
        promoted_to: PieceType,
        promoted_to_before: Bitboard,

    }
}

//01 function negamax(node, depth, color)
//02     if depth = 0 or node is a terminal node
//03         return color * the heuristic value of node
//
//04     bestValue := −∞
//05     foreach child of node
//06         v := −negamax(child, depth − 1, −color)
//07         bestValue := max( bestValue, v )
//08     return bestValue
pub fn search(game: &Game, depth: i32) -> Move {
    if depth == 0 {
        return Move::default();
    }

    let mut moves = generate_legal_moves(game, game.turn);
    //info!("mov list: {:?}", moves.iter().map(|x| x.to_string()).collect::<Vec<String>>());

    let mut rng = rand::thread_rng();

    /*moves.shuffle(&mut rng);

    // return random move
    if moves.len() >= 1 {
        return moves[0];
    }*/

    sort_moves(game, &mut moves);
    
    info!("moves: {:?}", moves.iter().map(|mv| mv.to_string()).collect::<Vec<String>>());

    let mut alpha = i32::min_value() + 1;
    let mut beta = i32::max_value();

    let mut valued_moves = moves.iter().map(|mov| {
        let new_game = apply_move(game, *mov);
        //info!("searching {}", mov.to_string());
        let val = -negamax(&new_game, -beta, -alpha, depth - 1);

        if val > -beta {
            //beta = -val;
            //info!("updated beta to {}", beta);
        }

        //info!("searched {} = {}", mov.to_string(), val);
        (*mov, -val)
    }).collect::<Vec<(Move, i32)>>();
    info!("movvalues: {:?}", valued_moves.iter().map(|mv| mv.0.to_string() + " - " + &mv.1.to_string()).collect::<Vec<String>>());

    valued_moves.sort_by_key(|mv| mv.1);


    if valued_moves.len() > 0 {
        let min_val = valued_moves[0].1;
        let best_moves = valued_moves.iter().filter(|mv| mv.1 <= min_val).map(|(mov, _)| *mov).collect::<Vec<Move>>();

        info!("bestmove value {}", -min_val);
        return best_moves[(rng.next_u64() % best_moves.len() as u64) as usize];
    }
    else {
        return Move::Default{ mov: SimpleMove{ from: 0, to: 0 }, piece_type: PAWN };
    }
}

fn negamax(game: &Game, mut alpha: i32, beta: i32, depth: i32) -> i32 {
    if depth == 0 {
        return quiescent_search(game, alpha, beta);
        let eval = evaluate(game);
        if eval != 0 {
            //info!("eval: {}", eval);
        }
        return eval;
    }

    const MIN_VALUE: i32 = i32::min_value() + 1;

    let mut best = MIN_VALUE;
    //let mut best_move = Move::default();
    //info!(" -> generate_legal_moves");

    if game.get_piece(KING, game.turn) == 0 { return MIN_VALUE; }

    let moves = generate_legal_moves(game, game.turn);

    if moves.len() == 0 {
        if is_check(game, game.turn) {
            // mate
            return MIN_VALUE;
        }
        else {
            // stalemate
            return 0;
        }
    }
    //info!(" -> generated_legal_moves {:?}", moves);
    for mov in moves {
        let new_game = apply_move(game, mov);
        //info!(" -> applied {} -> {}", mov.to_string(), depth - 1);
        let val = -negamax(&new_game, -beta, -alpha, depth - 1);
        //info!(" -> negamax returned {} -> {}", mov.to_string(), val);
        if val >= beta {
            return beta;
        }
        if val > alpha {
            alpha = val;
        }
        //info!(" -> negamaxed {} -> {}", mov.to_string(), depth - 1);
        if val > best {

            best = val;
            //best_move = mov;
        }
    }
    //info!("best alpha {}", best);
    return (alpha as f64 * 0.99) as i32;
}

fn quiescent_search(game: &Game, mut alpha: i32, beta: i32) -> i32 {
    const MIN_VALUE: i32 = i32::min_value() + 1;

    let val = evaluate(game);

    if val >= beta {
        return beta;
    }
    if val > alpha {
        alpha = val;
    }

    if game.get_piece(KING, game.turn) == 0 { return MIN_VALUE; }

    let moves = generate_attacking_moves(game, game.turn);
    for mov in moves {
        let new_game = apply_move(game, mov);
        let val = -quiescent_search(&new_game, -beta, -alpha);
        if val >= beta {
            return beta;
        }
        if val > alpha {
            alpha = val;
        }
    }
    return alpha;
}

pub fn apply_move(game: &Game, mov: Move) -> Game {
    let mut new_game = game.clone();

    let side = game.turn;
    let friends = game.get_all_side(side);
    let others = game.get_all_side(!side);
    match mov {
        Move::Default { mov, piece_type: pt } => {

            if pt == KING {
                // invalidate castling rights
                new_game.castling_rights[if game.turn == BLACK { 2 } else { 0 }] = false;
                new_game.castling_rights[if game.turn == BLACK { 3 } else { 1 }] = false;
            }

            if pt == ROOK {
                // invalidate castling rights
                if mov.from == square_from_indices(7, 7) {
                    new_game.castling_rights[0] = false;
                }
                else if mov.from == square_from_indices(0, 7) {
                    new_game.castling_rights[1] = false;
                }
                else if mov.from == square_from_indices(7, 7) {
                    new_game.castling_rights[2] = false;
                }
                else if mov.from == square_from_indices(7, 7) {
                    new_game.castling_rights[3] = false;
                }
            }

            if from_square(mov.to) | others != 0 {
                new_game.apply_mask(!from_square(mov.to));
            }
            let moved_piece = mov.apply_to(new_game.get_piece(pt, side));
            new_game.set_piece(pt, side, moved_piece);
        },
        Move::Castling { side, left } => {
            // invalidate future castling rights
            new_game.castling_rights[if game.turn == BLACK { 2 } else { 0 }] = false;
            new_game.castling_rights[if game.turn == BLACK { 3 } else { 1 }] = false;

            let king = game.get_piece(KING, side);
            let rook = if left {
                game.get_piece(ROOK, side) & (king << 4)
            }
            else {
                game.get_piece(ROOK, side) & (king >> 3)
            };

            let new_king = if left { king << 2 } else { king >> 2 };
            let new_rook = if left { rook >> 3 } else { rook << 2 };

            new_game.set_piece(KING, side, new_king);
            new_game.set_piece(ROOK, side, new_rook);
        },
        Move::EnPassant { mov, beaten } => {
            let pawns = game.pawns(side);
            if let Some(ep) = game.en_passant {
                *new_game.get_piece_mut(PAWN, side) &= !from_square(mov.from);
                *new_game.get_piece_mut(PAWN, side) |= from_square(mov.to);
                *new_game.get_piece_mut(PAWN, !side) |= from_square(beaten);
            }
        },
        Move::Promotion { mov, promote_to } => {
            if from_square(mov.to) | others != 0 {
                new_game.apply_mask(!from_square(mov.to));
            }
            new_game.apply_mask(!from_square(mov.from));

            match promote_to {
                QUEEN => { *new_game.queens_mut(side) |= from_square(mov.to); }
                ROOK => { *new_game.rooks_mut(side) |= from_square(mov.to); }
                BISHOP => { *new_game.bishops_mut(side) |= from_square(mov.to); }
                KNIGHT => { *new_game.knights_mut(side) |= from_square(mov.to); }
                _ => {
                    info!("internal error");
                }
            }
        },
    }

    new_game.turn = !new_game.turn;



    return new_game;
}
